#include <xmmspriv/xmms_plugin.h>
extern xmms_plugin_desc_t XMMS_PLUGIN_DESC_REPLAYGAIN;
extern xmms_plugin_desc_t XMMS_PLUGIN_DESC_EQUALIZER;
const xmms_plugin_desc_t *xmms_builtin_plugins[] = {
	&XMMS_PLUGIN_DESC_REPLAYGAIN,
	&XMMS_PLUGIN_DESC_EQUALIZER,
	NULL
};
