#ifndef __XMMS_FILE_BROWSE_H__
#define __XMMS_FILE_BROWSE_H__

#include <xmms/xmms_xformplugin.h>

gboolean xmms_file_browse (xmms_xform_t *xform, const gchar *url,
                           xmms_error_t *error);

#endif
